// 
//  MAC_MP3: A Low Energy Implementation of an Audio Decoder
//
//  Copyright (C) 2007 McMaster University
// 
//==============================================================================
// 
// This file is part of MAC_MP3
// 
// MAC_MP3 is distributed in the hope that it will be useful for further research,
// but WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
// or FITNESS FOR A PARTICULAR PURPOSE. MAC_MP3 is free; you can redistribute it
// and/or modify it provided that proper reference is provided to the authors. See
// the documents included in the "doc" folder for further details.
//
//==============================================================================


`timescale 1ns / 100ps
module ETH_interface(
   // Ethernet port
   TX_DATA_P,
   TX_ENABLE_P,
   TX_CLOCK_P,
   TX_ERROR_P,
   ENET_SLEW_P,
   RX_DATA_P,
   RX_DATA_VALID_P,
   RX_ERROR_P,
   RX_CLOCK_P,
   COLLISION_DETECTED_P,
   CARRIER_SENSE_P,
   PAUSE_P,
   MDIO_P,
   MDC_P,
   MDINIT_N,
   SSN_DATA_P,
   
   // System control
   resetn_in,
   clock_out,
   address_out,
   active_transmit_out,
   active_receive_out,
   error_out,
   
   // System transmit
   start_send_in,
   dest_MAC_in,
   type_in,
   range_begin_in,
   range_end_in,
   data_in,
   read_out,
   
   // System receive
   src_MAC_out,
   type_out,
   type_valid_out,
   range_begin_out,
   range_end_out,
   data_out,
   write_out
);

output   [3:0]    TX_DATA_P;           // Ethernet transmission data
output            TX_ENABLE_P;			// Ehternet transmission enable
input             TX_CLOCK_P;				// Ethernet transmission clock
input             TX_ERROR_P;				// Ethernet transmission error
output   [1:0]    ENET_SLEW_P;			// Ethernet slew settings
input    [3:0]    RX_DATA_P;				// Ethernet receive data
input             RX_DATA_VALID_P;		// Ethernet data valid
input             RX_ERROR_P;				// Ethernet receive error
input             RX_CLOCK_P;				// Ethernet receive clock
input             COLLISION_DETECTED_P;	// Ethernet collision detected
input             CARRIER_SENSE_P;		// Ethernet carrier sense
output            PAUSE_P;             // Ethernet pause
inout             MDIO_P;              // Ethernet config data
output            MDC_P;               // Ethernet config clock
input             MDINIT_N;            // Ethernet config init
inout             SSN_DATA_P;          // Silicon serial number access
   
input             resetn_in;
output            clock_out;
output   [23:0]   address_out;
output            active_transmit_out;
output            active_receive_out;
output            error_out;

input             start_send_in;
input    [47:0]   dest_MAC_in;
input    [15:0]   type_in;
input    [23:0]   range_begin_in;
input    [23:0]   range_end_in;
input    [31:0]   data_in;
output            read_out;
   
output   [47:0]   src_MAC_out;
output   [15:0]   type_out;
output            type_valid_out;
output   [23:0]   range_begin_out;
output   [23:0]   range_end_out;
output   [31:0]   data_out;
output            write_out;

reg      [23:0]   address_out;
reg               error_out;
reg      [31:0]   data_out;
reg               type_valid_out;
reg               write_out;
reg               read_out;
reg               active_transmit_out;
reg               active_receive_out;

wire     [3:0]    tx_data;
wire              tx_enable;
wire              tx_clock;
wire              tx_error;
wire     [1:0]    enet_slew;
wire     [3:0]    rx_data;
wire              rx_data_valid;
wire              rx_error;
wire              rx_clock;
wire              collision_detected;
wire              carrier_sense;
wire              pause;
wire              mdi, mdo, mdio_dir;
wire              mdc;
wire              mdinit;
wire              ssn_in, ssn_out, ssn_dir;

wire              clock;
reg      [3:0]    state;
reg               mode;
reg      [26:0]   state_counter;
reg      [47:0]   MAC_reg;
reg      [15:0]   type_reg;
reg      [23:0]   range_begin_reg, range_end_reg;
reg      [31:0]   buffer;
wire     [31:0]   buffer_next;
reg      [31:0]   CRC_reg;
wire     [31:0]   CRC_next, CRC_extend;

parameter ETH_RECV = 1'b0;
parameter ETH_SEND = 1'b1;

parameter ETH_WAIT = 4'h0;
parameter ETH_SYNC = 4'h1;
parameter ETH_DEST = 4'h2;
parameter ETH_SRC = 4'h3;
parameter ETH_TYPE = 4'h4;
parameter ETH_RANGE_BEGIN = 4'h5;
parameter ETH_RANGE_END = 4'h6;
parameter ETH_DATA = 4'h7;
parameter ETH_CRC = 4'h8;
parameter ETH_MAC_ADDRESS = 48'h5F43D4EA7B13;
//parameter ETH_MAC_ADDRESS = 48'h000103df9121;
wire     [47:0]   MAC_ADDRESS;

assign MAC_ADDRESS = ETH_MAC_ADDRESS;
assign clock_out = clock;

assign src_MAC_out = MAC_reg;
assign type_out = type_reg;
assign range_begin_out = range_begin_reg;
assign range_end_out = range_end_reg;
assign tx_enable = active_transmit_out;
assign tx_data = buffer[27:24];
assign buffer_next = {  buffer[19:16],
                        buffer[31:28],
                        buffer[11:8],
                        buffer[23:20],
                        buffer[3:0],
                        buffer[15:12],
                        rx_data,
                        buffer[7:4]};

always @(posedge clock or negedge resetn_in)
begin
   if (~resetn_in)
   begin
      state <= ETH_WAIT;
      mode <= ETH_RECV;
      state_counter <= 27'd0;
      MAC_reg <= 48'h000000000000;
      type_reg <= 16'h0000;
      range_begin_reg <= 24'h000000;
//range_end_reg <= 24'h1FFFFF;
range_end_reg <= range_end_reg;
//      range_end_reg <= 24'h000000;
      buffer <= 32'h00000000;
      CRC_reg <= 32'h00000000;
      error_out <= 1'b0;
      data_out <= 32'h00000000;
      address_out <= 24'h000000;
      read_out <= 1'b0;
      write_out <= 1'b0;
      active_receive_out <= 1'b0;
      active_transmit_out <= 1'b0;
      type_valid_out <= 1'b0;
   end
   else
   begin #10
      state_counter <= state_counter + 1;
      buffer <= buffer_next;
      CRC_reg <= CRC_next;
      type_valid_out <= 1'b0;
      case (state)
         ETH_WAIT        : begin
						buffer <= 32'h00000000;
                              active_receive_out <= 1'b0;
                              active_transmit_out <= 1'b0;
                              if (!(tx_error || rx_error))
                                 if ((rx_data_valid == 1'b1) && (rx_data == 4'h5))
                                 begin
                                    state <= ETH_SYNC;
                                    mode <= ETH_RECV;
                                    error_out <= 1'b0;
                                    active_receive_out <= 1'b1;
                                 end
                                 else if (start_send_in && !carrier_sense)
                                 begin
                                    state <= ETH_SYNC;
                                    mode <= ETH_SEND;
                                    state_counter <= 27'd0; 
                                    MAC_reg <= dest_MAC_in;
                                    type_reg <= type_in;
                                    range_begin_reg <= range_begin_in;
                                    range_end_reg <= range_end_in;
                                    error_out <= 1'b0;
                                    buffer <= 32'h55555555;
                                    active_transmit_out <= 1'b1;
                                 end
                           end
         ETH_SYNC        : if ((mode == ETH_SEND) && (state_counter == 27'd7))
                              buffer <= 32'h555555D5;
                           else if (((rx_data == 4'hD) && (mode == ETH_RECV)) || 
                               ((state_counter == 27'd15) && (mode == ETH_SEND)))
                           begin
                              state <= ETH_DEST;
                              state_counter <= 27'd0;
                              buffer <= MAC_reg[47:16];
                              CRC_reg <= 32'h00000000;
                           end
                           else if ((mode == ETH_RECV) && (rx_data != 4'h5))
                              state <= ETH_WAIT;
         ETH_DEST        : begin
                              if (state_counter == 27'd7) 
                                 if (mode == ETH_RECV)
                                    MAC_reg[47:16] <= buffer_next;
                                 else
                                    buffer[31:16] <= MAC_reg[15:0];
                              
                              if (state_counter == 27'd11)
                                 if ((mode == ETH_SEND) ||
                                     ((mode == ETH_RECV) && 
                                      ({MAC_reg[47:16],buffer_next[15:0]} == MAC_ADDRESS)))
                                 begin
                                    state <= ETH_SRC;
                                    state_counter <= 27'd0;
                                    buffer[31:0] <= MAC_ADDRESS[47:16];
                                    if (mode == ETH_RECV)
                                       MAC_reg[15:0] <= buffer_next[15:0];
                                 end
                                 else
                                    state <= ETH_WAIT;
                           end
         ETH_SRC         : begin
                              if (state_counter == 27'd7)
                                 if (mode == ETH_RECV)
                                    MAC_reg[47:16] <= buffer_next;
                                 else
                                    buffer[31:16] <= MAC_ADDRESS[15:0];

                              if (state_counter == 27'd11)
                              begin
                                 state <= ETH_TYPE;
                                 state_counter <= 27'd0;
                                 buffer[31:16] <= type_reg;
                                 if (mode == ETH_RECV)
                                    MAC_reg[15:0] <= buffer_next[15:0];
                              end
                           end
         ETH_TYPE        : if (state_counter == 27'd3)
                           begin
                              state <= ETH_RANGE_BEGIN;
                              state_counter <= 27'd0;
                              buffer[31:8] <= range_begin_reg;
                              if (mode == ETH_RECV)
                              begin
                                 type_reg <= buffer_next[15:0];
                                 type_valid_out <= 1'b1;
                              end
                           end
         ETH_RANGE_BEGIN : if ((state_counter == 27'd4) && (mode == ETH_SEND))
	                      begin
                              address_out <= range_begin_reg;
						read_out <= 1'b1;
                           end
                           else if (state_counter == 27'd5)
                           begin
                              read_out <= 1'b0;
                              state <= ETH_RANGE_END;
                              state_counter <= 27'd0;
                              buffer[31:8] <= range_end_reg;
                              if (mode == ETH_RECV)
                                 range_begin_reg <= buffer_next[23:0];
                           end
         ETH_RANGE_END   : if (state_counter == 27'd5)
                           begin
                              state <= ETH_DATA;
                              buffer <= data_in;
                              state_counter <= {range_begin_reg, 3'h0};
                              if (mode == ETH_RECV)
                                 range_end_reg <= buffer_next[23:0];
                           end
         ETH_DATA        : begin 
                              write_out <= 1'b0;
                              read_out <= 1'b0;
                              
                              if (state_counter[2:0] == 3'h7)
                              begin
                                 if (mode == ETH_RECV)
                                 begin
                                    write_out <= 1'b1;
                                    address_out <= state_counter[26:3];
                                    data_out <= buffer_next;
                                 end
                                 else
                                    buffer <= data_in;
                                 
                                 if (state_counter[26:3] == range_end_reg)
                                 begin
                                    state <= ETH_CRC;
                                    state_counter <= 27'd0;
                                    CRC_reg <= CRC_extend;
                                    buffer <= CRC_extend;
                                 end
                              end
                              else if ((state_counter[2:0] == 3'h0) && (mode == ETH_SEND))
                              begin
                                 address_out <= state_counter[26:3] + 1;
                                 if (state_counter[26:3] != range_end_reg)
                                    read_out <= 1'b1;
                              end
                           end
         ETH_CRC         : begin
                              read_out <= 1'b0;
                              write_out <= 1'b0;
                              CRC_reg <= CRC_reg;
                              if (state_counter == 27'd7)
                              begin
                                 state <= ETH_WAIT;
                                 if ((mode == ETH_RECV) && (CRC_reg != buffer_next))
                                    error_out <= 1'b1;
                              end
                           end
      endcase
   end
end

// CRC generation
wire     [3:0]    CRC_source_data, CRC_data;
assign CRC_source_data = (mode == ETH_RECV) ? rx_data : tx_data;
assign CRC_data = ((state == ETH_DEST) && (state_counter <= 7)) ? ~CRC_source_data : CRC_source_data;
next_CRC CRC_gen (.curr_CRC(CRC_reg), .data(CRC_data), .next_CRC(CRC_next));

wire     [31:0]   temp_CRC1, temp_CRC2, temp_CRC3, temp_CRC4, temp_CRC5, temp_CRC6, temp_CRC7, temp_CRC8;
next_CRC CRC1 (.curr_CRC(CRC_next), .data(4'h0), .next_CRC(temp_CRC1));
next_CRC CRC2 (.curr_CRC(temp_CRC1), .data(4'h0), .next_CRC(temp_CRC2));
next_CRC CRC3 (.curr_CRC(temp_CRC2), .data(4'h0), .next_CRC(temp_CRC3));
next_CRC CRC4 (.curr_CRC(temp_CRC3), .data(4'h0), .next_CRC(temp_CRC4));
next_CRC CRC5 (.curr_CRC(temp_CRC4), .data(4'h0), .next_CRC(temp_CRC5));
next_CRC CRC6 (.curr_CRC(temp_CRC5), .data(4'h0), .next_CRC(temp_CRC6));
next_CRC CRC7 (.curr_CRC(temp_CRC6), .data(4'h0), .next_CRC(temp_CRC7));
next_CRC CRC8 (.curr_CRC(temp_CRC7), .data(4'h0), .next_CRC(temp_CRC8));
assign CRC_extend = ~{  temp_CRC8[24], temp_CRC8[25], temp_CRC8[26], temp_CRC8[27],
                        temp_CRC8[28], temp_CRC8[29], temp_CRC8[30], temp_CRC8[31],
                        temp_CRC8[16], temp_CRC8[17], temp_CRC8[18], temp_CRC8[19], 
                        temp_CRC8[20], temp_CRC8[21], temp_CRC8[22], temp_CRC8[23], 
                        temp_CRC8[8], temp_CRC8[9], temp_CRC8[10], temp_CRC8[11], 
                        temp_CRC8[12], temp_CRC8[13], temp_CRC8[14], temp_CRC8[15], 
                        temp_CRC8[0], temp_CRC8[1], temp_CRC8[2], temp_CRC8[3],
                        temp_CRC8[4], temp_CRC8[5], temp_CRC8[6], temp_CRC8[7]}; 

assign enet_slew = 2'b00;
assign pause = 1'b0;
assign mdc = 1'b0;
assign mdo = 1'b0; assign mdio_dir = 1'b0;
assign ssn_out = 1'b0; assign ssn_dir = 1'b0;

// Buffering
OBUF TX_DATA3_BUF (.I(tx_data[3]), .O(TX_DATA_P[3]));
OBUF TX_DATA2_BUF (.I(tx_data[2]), .O(TX_DATA_P[2]));
OBUF TX_DATA1_BUF (.I(tx_data[1]), .O(TX_DATA_P[1]));
OBUF TX_DATA0_BUF (.I(tx_data[0]), .O(TX_DATA_P[0]));
OBUF TX_ENABLE_BUF (.I(tx_enable), .O(TX_ENABLE_P));
OBUF ENET_SLEW1_BUF	(.I(enet_slew[1]), .O(ENET_SLEW_P[1]));
OBUF ENET_SLEW0_BUF	(.I(enet_slew[0]), .O(ENET_SLEW_P[0]));
OBUF PAUSE_BUF	(.I(pause), .O(PAUSE_P));
OBUF MDC_BUF (.I(mdc), .O(MDC_P));

IBUF TX_CLOCK_BUF (.I(TX_CLOCK_P), .O(tx_clock));
IBUF TX_ERROR_BUF (.I(TX_ERROR_P), .O(tx_error));
IBUF RX_DATA3_BUF (.I(RX_DATA_P[3]), .O(rx_data[3]));
IBUF RX_DATA2_BUF (.I(RX_DATA_P[2]), .O(rx_data[2]));
IBUF RX_DATA1_BUF (.I(RX_DATA_P[1]), .O(rx_data[1]));
IBUF RX_DATA0_BUF (.I(RX_DATA_P[0]), .O(rx_data[0]));
IBUF RX_DATA_VALID_BUF (.I(RX_DATA_VALID_P), .O(rx_data_valid));
IBUF RX_ERROR_BUF (.I(RX_ERROR_P), .O(rx_error));
IBUF RX_CLOCK_BUF (.I(RX_CLOCK_P), .O(rx_clock));
IBUF COLLISION_DETECTED_BUF (.I(COLLISION_DETECTED_P), .O(collision_detected));
IBUF CARRIER_SENSE_BUF (.I(CARRIER_SENSE_P), .O(carrier_sense));
IBUF MDINIT_BUF (.I(MDINIT_N), .O(mdinit));

IOBUF MDIO_BUF (.I(mdi), .O(mdo), .T(mdio_dir), .IO(MDIO_P));
IOBUF SSN_DATA_BUF (.I(ssn_in), .O(ssn_out), .T(ssn_dir), .IO(SSN_DATA_P));

BUFG rx_clock_gbuf (.I(rx_clock), .O(clock));
endmodule
